@@ -6,7 +6,8 @@ import shutil
 from datetime import datetime, timezone
 from textwrap import TextWrapper
 from functools import lru_cache
-from typing import List, Iterable, Union, Optional
+from typing import List, Dict, Iterable, Union, Optional, TextIO
+from .jc_types import TimeStampFormatType
 
 
 def _asciify(string: str) -> str:
@@ -21,7 +22,13 @@ def _asciify(string: str) -> str:
     return string
 
 
-def _safe_print(string: str, sep=' ', end='\n', file=sys.stdout, flush=False) -> None:
+def _safe_print(
+    string: str,
+    sep: str = ' ',
+    end: str = '\n',
+    file: TextIO = sys.stdout,
+    flush: bool = False
+) -> None:
     """Output for both UTF-8 and ASCII encoding systems"""
     try:
         print(string, sep=sep, end=end, file=file, flush=flush)
@@ -29,10 +36,19 @@ def _safe_print(string: str, sep=' ', end='\n', file=sys.stdout, flush=False) ->
         print(_asciify(string), sep=sep, end=end, file=file, flush=flush)
 
 
+def _safe_pager(string: str) -> None:
+    """Pager output for both UTF-8 and ASCII encoding systems"""
+    from pydoc import pager
+    try:
+        pager(string)
+    except UnicodeEncodeError:
+        pager(_asciify(string))
+
+
 def warning_message(message_lines: List[str]) -> None:
     """
-    Prints warning message for non-fatal issues. The first line is
-    prepended with 'jc:  Warning - ' and subsequent lines are indented.
+    Prints warning message to `STDERR` for non-fatal issues. The first line
+    is prepended with 'jc:  Warning - ' and subsequent lines are indented.
     Wraps text as needed based on the terminal width.
 
     Parameters:
@@ -67,7 +83,7 @@ def warning_message(message_lines: List[str]) -> None:
 
 def error_message(message_lines: List[str]) -> None:
     """
-    Prints an error message for fatal issues. The first line is
+    Prints an error message to `STDERR` for fatal issues. The first line is
     prepended with 'jc:  Error - ' and subsequent lines are indented.
     Wraps text as needed based on the terminal width.
 
@@ -97,10 +113,25 @@ def error_message(message_lines: List[str]) -> None:
         _safe_print(message, file=sys.stderr)
 
 
-def compatibility(mod_name: str, compatible: List, quiet: bool = False) -> None:
+def is_compatible(compatible: List[str]) -> bool:
     """
-    Checks for the parser's compatibility with the running OS
-    platform.
+    Returns True if the parser is compatible with the running OS platform.
+    """
+    platform_found = False
+
+    for platform in compatible:
+        if sys.platform.startswith(platform):
+            platform_found = True
+            break
+
+    return platform_found
+
+
+def compatibility(mod_name: str, compatible: List[str], quiet: bool = False) -> None:
+    """
+    Checks for the parser's compatibility with the running OS platform and
+    prints a warning message to `STDERR` if not compatible and
+    `quiet=False.`
 
     Parameters:
 
@@ -116,41 +147,39 @@ def compatibility(mod_name: str, compatible: List, quiet: bool = False) -> None:
 
         None - just prints output to STDERR
     """
-    if not quiet:
-        platform_found = False
-
-        for platform in compatible:
-            if sys.platform.startswith(platform):
-                platform_found = True
-                break
-
-        if not platform_found:
-            mod = mod_name.split('.')[-1]
-            compat_list = ', '.join(compatible)
-            warning_message([
-                f'{mod} parser is not compatible with your OS ({sys.platform}).',
-                f'Compatible platforms: {compat_list}'
-            ])
+    if not quiet and not is_compatible(compatible):
+        mod = mod_name.split('.')[-1]
+        compat_list = ', '.join(compatible)
+        warning_message([
+            f'{mod} parser is not compatible with your OS ({sys.platform}).',
+            f'Compatible platforms: {compat_list}'
+        ])
 
 
-def has_data(data: str) -> bool:
+def has_data(data: Union[str, bytes]) -> bool:
     """
-    Checks if the input contains data. If there are any non-whitespace
-    characters then return True, else return False.
+    Checks if the string input contains data. If there are any
+    non-whitespace characters then return `True`, else return `False`.
+
+    For bytes, returns True if there is any data.
 
     Parameters:
 
-        data:        (string) input to check whether it contains data
+        data:        (string, bytes) input to check whether it contains data
 
     Returns:
 
         Boolean      True if input string (data) contains non-whitespace
-                     characters, otherwise False
+                     characters, otherwise False. For bytes data, returns
+                     True if there is any data, otherwise False.
     """
-    return bool(data and not data.isspace())
+    if isinstance(data, str):
+        return bool(data and not data.isspace())
+
+    return bool(data)
 
 
-def convert_to_int(value: Union[str, float]) -> Optional[int]:
+def convert_to_int(value: object) -> Optional[int]:
     """
     Converts string and float input to int. Strips all non-numeric
     characters from strings.
@@ -180,7 +209,7 @@ def convert_to_int(value: Union[str, float]) -> Optional[int]:
         return None
 
 
-def convert_to_float(value: Union[str, int]) -> Optional[float]:
+def convert_to_float(value: object) -> Optional[float]:
     """
     Converts string and int input to float. Strips all non-numeric
     characters from strings.
@@ -206,7 +235,7 @@ def convert_to_float(value: Union[str, int]) -> Optional[float]:
         return None
 
 
-def convert_to_bool(value: Union[str, int, float]) -> bool:
+def convert_to_bool(value: object) -> bool:
     """
     Converts string, integer, or float input to boolean by checking
     for 'truthy' values.
@@ -251,9 +280,11 @@ def input_type_check(data: str) -> None:
 
 
 class timestamp:
+    __slots__ = ('string', 'format', 'naive', 'utc')
+
     def __init__(self,
-                 datetime_string: str,
-                 format_hint: Optional[Iterable] = None
+                 datetime_string: Optional[str],
+                 format_hint: Optional[Iterable[int]] = None
     ) -> None:
         """
         Input a datetime text string of several formats and convert to a
@@ -296,12 +327,15 @@ class timestamp:
         self.naive = dt['timestamp_naive']
         self.utc = dt['timestamp_utc']
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f'timestamp(string={self.string!r}, format={self.format}, naive={self.naive}, utc={self.utc})'
 
     @staticmethod
     @lru_cache(maxsize=512)
-    def _parse_dt(dt_string, format_hint=None):
+    def _parse_dt(
+        dt_string: Optional[str],
+        format_hint: Optional[Iterable[int]] = None
+    ) -> Dict[str, Optional[int]]:
         """
         Input a datetime text string of several formats and convert to
         a naive or timezone-aware epoch timestamp in UTC.
@@ -348,40 +382,15 @@ class timestamp:
 
                 If the conversion completely fails, all fields will be None.
         """
-        data = dt_string or ''
-        normalized_datetime = ''
-        utc_tz = False
-        dt = None
-        dt_utc = None
-        timestamp_naive = None
-        timestamp_utc = None
-        timestamp_obj = {
-            'format': None,
-            'timestamp_naive': None,
-            'timestamp_utc': None
-        }
-        utc_tz = False
-
-        # convert format_hint to a tuple so it is hashable (for lru_cache)
-        if not format_hint:
-            format_hint = tuple()
-        else:
-            format_hint = tuple(format_hint)
-
-        # sometimes UTC is referenced as 'Coordinated Universal Time'. Convert to 'UTC'
-        data = data.replace('Coordinated Universal Time', 'UTC')
-
-        if 'UTC' in data:
-            utc_tz = True
-            if 'UTC+' in data or 'UTC-' in data:
-                utc_tz = bool('UTC+0000' in data or 'UTC-0000' in data)
-
-        elif '+0000' in data or '-0000' in data:
-            utc_tz = True
-
-        formats = [
+        formats: tuple[TimeStampFormatType, ...] = (
             {'id': 1000, 'format': '%a %b %d %H:%M:%S %Y', 'locale': None},  # manual C locale format conversion: Tue Mar 23 16:12:11 2021 or Tue Mar 23 16:12:11 IST 2021
             {'id': 1100, 'format': '%a %b %d %H:%M:%S %Y %z', 'locale': None}, # git date output: Thu Mar 5 09:17:40 2020 -0800
+            {'id': 1300, 'format': '%Y-%m-%dT%H:%M:%S.%f%Z', 'locale': None}, # ISO Format with UTC (found in syslog 5424): 2003-10-11T22:14:15.003Z
+            {'id': 1310, 'format': '%Y-%m-%dT%H:%M:%S.%f', 'locale': None}, # ISO Format without TZ (found in syslog 5424): 2003-10-11T22:14:15.003
+            {'id': 1400, 'format': '%b %d %Y %H:%M:%S.%f UTC', 'locale': None}, # CEF Format with UTC: Nov 08 2022 12:30:00.111 UTC
+            {'id': 1410, 'format': '%b %d %Y %H:%M:%S.%f', 'locale': None}, # CEF Format without TZ: Nov 08 2022 12:30:00.111
+            {'id': 1420, 'format': '%b %d %Y %H:%M:%S UTC', 'locale': None}, # CEF Format with UTC without microseconds: Nov 08 2022 12:30:00 UTC
+            {'id': 1430, 'format': '%b %d %Y %H:%M:%S', 'locale': None}, # CEF Format without TZ or microseconds: Nov 08 2022 12:30:00
             {'id': 1500, 'format': '%Y-%m-%d %H:%M', 'locale': None},  # en_US.UTF-8 local format (found in who cli output): 2021-03-23 00:14
             {'id': 1600, 'format': '%m/%d/%Y %I:%M %p', 'locale': None},  # Windows english format (found in dir cli output): 12/07/2019 02:09 AM
             {'id': 1700, 'format': '%m/%d/%Y, %I:%M:%S %p', 'locale': None},  # Windows english format wint non-UTC tz (found in systeminfo cli output): 3/22/2021, 1:15:51 PM (UTC-0600)
@@ -404,11 +413,11 @@ class timestamp:
             {'id': 8200, 'format': '%A %d %B %Y, %H:%M:%S UTC%z', 'locale': ''},  # fr_FR.utf8 locale format (found in date cli output): vendredi 26 mars 2021, 13:26:46 (UTC+0000)
             {'id': 8300, 'format': '%A %d %B %Y, %H:%M:%S', 'locale': ''},  # fr_FR.utf8 locale format with non-UTC tz (found in date cli output): vendredi 26 mars 2021, 13:26:46 (UTC-0400)
             {'id': 9000, 'format': '%c', 'locale': ''}  # locally configured locale format conversion: Could be anything :) this is a last-gasp attempt
-        ]
+        )
 
         # from https://www.timeanddate.com/time/zones/
         # only removed UTC timezone and added known non-UTC offsets
-        tz_abbr = [
+        tz_abbr: set[str] = {
             'A', 'ACDT', 'ACST', 'ACT', 'ACWST', 'ADT', 'AEDT', 'AEST', 'AET', 'AFT', 'AKDT',
             'AKST', 'ALMT', 'AMST', 'AMT', 'ANAST', 'ANAT', 'AQTT', 'ART', 'AST', 'AT', 'AWDT',
             'AWST', 'AZOST', 'AZOT', 'AZST', 'AZT', 'AoE', 'B', 'BNT', 'BOT', 'BRST', 'BRT', 'BST',
@@ -428,29 +437,77 @@ class timestamp:
             'TOT', 'TRT', 'TVT', 'U', 'ULAST', 'ULAT', 'UYST', 'UYT', 'UZT', 'V', 'VET', 'VLAST',
             'VLAT', 'VOST', 'VUT', 'W', 'WAKT', 'WARST', 'WAST', 'WAT', 'WEST', 'WET', 'WFT',
             'WGST', 'WGT', 'WIB', 'WIT', 'WITA', 'WST', 'WT', 'X', 'Y', 'YAKST', 'YAKT', 'YAPT',
-            'YEKST', 'YEKT', 'Z', 'UTC-1200', 'UTC-1100', 'UTC-1000', 'UTC-0930', 'UTC-0900',
+            'YEKST', 'YEKT', 'UTC-1200', 'UTC-1100', 'UTC-1000', 'UTC-0930', 'UTC-0900',
             'UTC-0800', 'UTC-0700', 'UTC-0600', 'UTC-0500', 'UTC-0400', 'UTC-0300', 'UTC-0230',
             'UTC-0200', 'UTC-0100', 'UTC+0100', 'UTC+0200', 'UTC+0300', 'UTC+0400', 'UTC+0430',
             'UTC+0500', 'UTC+0530', 'UTC+0545', 'UTC+0600', 'UTC+0630', 'UTC+0700', 'UTC+0800',
             'UTC+0845', 'UTC+0900', 'UTC+1000', 'UTC+1030', 'UTC+1100', 'UTC+1200', 'UTC+1300',
             'UTC+1345', 'UTC+1400'
-        ]
+        }
+
+        offset_suffixes: tuple[str, ...] = (
+            '-12:00', '-11:00', '-10:00', '-09:30', '-09:00',
+            '-08:00', '-07:00', '-06:00', '-05:00', '-04:00', '-03:00', '-02:30',
+            '-02:00', '-01:00', '+01:00', '+02:00', '+03:00', '+04:00', '+04:30',
+            '+05:00', '+05:30', '+05:45', '+06:00', '+06:30', '+07:00', '+08:00',
+            '+08:45', '+09:00', '+10:00', '+10:30', '+11:00', '+12:00', '+13:00',
+            '+13:45', '+14:00'
+        )
+
+        data: str = dt_string or ''
+        normalized_datetime: str = ''
+        utc_tz: bool = False
+        dt: Optional[datetime] = None
+        dt_utc: Optional[datetime] = None
+        timestamp_naive: Optional[int] = None
+        timestamp_utc: Optional[int] = None
+        timestamp_obj: Dict[str, Optional[int]] = {
+            'format': None,
+            'timestamp_naive': None,
+            'timestamp_utc': None
+        }
+
+        # convert format_hint to a tuple so it is hashable (for lru_cache)
+        if not format_hint:
+            format_hint = tuple()
+        else:
+            format_hint = tuple(format_hint)
+
+        # sometimes UTC is referenced as 'Coordinated Universal Time'. Convert to 'UTC'
+        data = data.replace('Coordinated Universal Time', 'UTC')
+
+        # UTC can also be indicated with 'Z' for Zulu time (ISO-8601). Convert to 'UTC'
+        data = data.replace('Z', 'UTC')
+
+        if 'UTC' in data:
+            utc_tz = True
+            if 'UTC+' in data or 'UTC-' in data:
+                utc_tz = bool('UTC+0000' in data or 'UTC-0000' in data)
+
+        elif '+0000' in data or '-0000' in data:
+            utc_tz = True
 
         # normalize the timezone by taking out any timezone reference, except UTC
         cleandata = data.replace('(', '').replace(')', '')
-        normalized_datetime_list = []
+        normalized_datetime_list: List[str] = []
         for term in cleandata.split():
             if term not in tz_abbr:
                 normalized_datetime_list.append(term)
 
         normalized_datetime = ' '.join(normalized_datetime_list)
 
+        # remove non UTC offset suffixes at the end of the string
+        for suffix in offset_suffixes:
+            if normalized_datetime.endswith(suffix):
+                normalized_datetime = normalized_datetime[0:-len(suffix)]
+                break
+
         # normalize further by converting any greater-than 6-digit subsecond to 6-digits
         p = re.compile(r'(\W\d\d:\d\d:\d\d\.\d{6})\d+\W')
         normalized_datetime = p.sub(r'\g<1> ', normalized_datetime)
 
         # try format hints first, then fall back to brute-force method
-        hint_obj_list = []
+        hint_obj_list: List[TimeStampFormatType] = []
         for fmt_id in format_hint:
             for fmt in formats:
                 if fmt_id == fmt['id']:
